import { BaseConnection, WsConnection } from "tsrpc";
import { server } from "../..";
import { RandomEx } from "../../Engine/Utils/Number/RandomEx";
import { EGameState } from "../../shared/protocols/define/enum";
import { MsgChangeState } from "../../shared/protocols/room/MsgChangeState";
import { ServiceType } from "../../shared/protocols/serviceProto";
import { RoomConnData } from "./RoomUtils";


/**
 * Room
 */
export default class Room {

	//#region static

	private static maxRoomId: number = 0;
	private static rooms: { [roomId: number]: Room } = {};

	//#endregion

	//#region public

	public RoomId: number = 0;

	//#endregion

	//#region private

	private conns: BaseConnection[] = [];

	private connData: Map<BaseConnection<any>, RoomConnData> = new Map<BaseConnection, RoomConnData>();

	//#endregion

	//#region Lifecycle

	/**
	 *
	 */
	constructor() {
		// 给每个新房间生成一个唯一的 ID
		this.RoomId = ++Room.maxRoomId;
		Room.rooms[this.RoomId] = this;
	}

	//#endregion

	//#region Custom

	/** GetRoom */
	public static GetRoom(roomId: number): Room {
		return this.rooms[roomId];
	}

	//#endregion

	//#region API

	/** Exit */
	public Exit(conn: BaseConnection): void {
		server.broadcastMsg("room/Exit", 0, <WsConnection<ServiceType>[]>this.conns);
	}

	/** GotoGame */
	public GotoGame(): void {
		server.broadcastMsg("room/GoToGame", 0, <WsConnection<ServiceType>[]>this.conns);
	}

	/** ChangeState */
	public ChangeState(needData: boolean = false, state?: EGameState, conn?: BaseConnection): void {
		this.connData.forEach((connData: RoomConnData, key: BaseConnection) => {
			let target: WsConnection<ServiceType>[] = [];
			let data: MsgChangeState = [];
			if (conn) {
				if (state) {
					target = <WsConnection<ServiceType>[]>this.conns.filter((conn) => conn === key);
					const apiNum: number = conn === key ? 100 : 200;
					data = [state + apiNum];
				} else if (conn) {
					target = <WsConnection<ServiceType>[]>this.conns.filter((conn) => conn === key);
					const apiNum: number = conn === key ? 100 : 200;
					data = [(connData.state ?? EGameState.Ready) + apiNum];
				}
			} else {
				target = <WsConnection<ServiceType>[]>this.conns.filter((conn) => conn === key);
				data = [connData.state];
			}
			if (needData) {
				data.push(connData.card);
			}
			server.broadcastMsg("room/ChangeState", data, target);
		});
		switch (state) {
			case EGameState.Ready:
				this.Ready(conn);
				break;

			case EGameState.TurnEnd:
				this.TurnEnd(conn);
				break;

			default:
				break;
		}
	}

	//#endregion

	//#region ChangeState

	/** Ready */
	public Ready(conn: BaseConnection): void {
		const self: this = this;
		const cardCount: number = 24;
		const data: RoomConnData = this.connData.get(conn);
		data.state = EGameState.Ready;
		this.connData.set(conn, data);
		let canStart: boolean = true;
		this.connData.forEach((connData: RoomConnData) => {
			if (connData.state === undefined || connData.state !== EGameState.Ready) {
				canStart = false;
			}
		});
		if (canStart) {
			const now: number = RandomEx.GetInt(0, 2);
			this.connData.forEach((connData: RoomConnData, key: BaseConnection) => {
				const card: number = RandomEx.GetInt(0, cardCount);
				connData.state = key === self.conns[now] ? EGameState.MyTurnCards : EGameState.MyWait;
				connData.card = card;
				self.connData.set(key, connData);
			});
			this.ChangeState(true);
		}
	}

	/** TurnEnd */
	public TurnEnd(conn: BaseConnection): void {
		const self: this = this;
		this.connData.forEach((connData: RoomConnData, key: BaseConnection) => {
			connData.state = EGameState.MyWait;
			self.connData.set(key, connData);
		});
		const index: number = this.conns.indexOf(conn);
		const nextIndex: number = (index + 1) % this.conns.length;
		const nextConn: BaseConnection = this.conns[nextIndex];
		const nextData: RoomConnData = this.connData.get(nextConn);
		nextData.state = EGameState.MyTurnCards;
		this.connData.set(nextConn, nextData);
		this.ChangeState();
	}

	//#endregion

	//#region Custom

	/** Join */
	public Join(conn: BaseConnection): void {
		this.conns.push(conn);
		const data: RoomConnData = {
			state: undefined,
			card: undefined,
		};
		this.connData.set(conn, data);
	}

	/** RestartGame */
	public RestartGame(): void {

	}

	/** ConnCount */
	public ConnCount(): number {
		return this.conns.length;
	}

	//#endregion
}
